models = {
    "openai": {
        "gpt-4.1": {
            "name": "gpt-4.1",
            "api_keys": ["openai_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "gpt-4.1-mini": {
            "name": "gpt-4.1-mini",
            "api_keys": ["openai_api_key"],
            "speed": "medium",
            "accuracy": "medium",
        },
        "gpt-4.1-nano": {
            "name": "gpt-4.1-nano",
            "api_keys": ["openai_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "anthropic": {
        "claude-sonnet-4-20250514": {
            "name": "claude-sonnet-4-20250514",
            "api_keys": ["anthropic_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-7-sonnet-20250219": {
            "name": "claude-3-7-sonnet-20250219",
            "api_keys": ["anthropic_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-5-haiku-20241022": {
            "name": "claude-3-5-haiku-20241022",
            "api_keys": ["anthropic_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/openai": {
        "gpt-4.1": {
            "name": "gpt-4.1",
            "api_keys": ["openai_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "gpt-4.1-mini": {
            "name": "gpt-4.1-mini",
            "api_keys": ["openai_api_key"],
            "speed": "medium",
            "accuracy": "medium",
        },
        "gpt-4.1-nano": {
            "name": "gpt-4.1-nano",
            "api_keys": ["openai_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/anthropic": {
        "claude-sonnet-4": {
            "name": "claude-sonnet-4",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-7-sonnet": {
            "name": "claude-3-7-sonnet",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-5-haiku": {
            "name": "claude-3-5-haiku",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/google": {
        "gemini-2.5-flash": {
            "name": "gemini-2.5-flash",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-flash-lite": {
            "name": "gemini-2.5-flash-lite",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
        "gemini-2.0-flash-001": {
            "name": "gemini-2.0-flash-001",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-pro": {
            "name": "gemini-2.5-pro",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
    },
    "gemini": {
        "gemini-2.5-flash": {
            "name": "gemini-2.5-flash",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-flash-lite": {
            "name": "gemini-2.5-flash-lite",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
        "gemini-2.0-flash-001": {
            "name": "gemini-2.0-flash-001",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-pro": {
            "name": "gemini-2.5-pro",
            "api_keys": ["gemini_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
    },
}
